///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hdate_time;

import 'package:haystack/src/string_utils.dart';
import 'package:quiver/core.dart';

import 'HDate.dart';
import 'HTime.dart';
import 'HTimeZone.dart';
import 'HVal.dart';
import 'ParseError.dart';

/// HDateTime models a timestamp with a specific timezone.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
class HDateTime extends HVal {
  factory HDateTime.fromValues(
      int year, int month, int day, int hour, int min, int sec, HTimeZone tz) {
    HDate date = new HDate(year, month, day);
    HTime time = new HTime(hour, min, sec);
    if (tz == null) tz = HTimeZone.DEFAULT;
    return new HDateTime._(date, time, tz);
  }

  factory HDateTime.fromMillis(int ms, [HTimeZone tz]) {
    return new HDateTime.fromDateTime(
        new DateTime.fromMillisecondsSinceEpoch(ms, isUtc: true), tz);
  }

  factory HDateTime.fromDateTime(DateTime dt, [HTimeZone tz]) {
    if (tz == null) tz = HTimeZone.DEFAULT;
    DateTime fixedTime = dt.add(new Duration(seconds: tz.offset));
    return new HDateTime(new HDate.fromDateTime(fixedTime),
        new HTime.fromDateTime(fixedTime), tz);
  }

  factory HDateTime.fromString(String s) {
    // date
    int tIdx = s.indexOf('T');
    if (tIdx < 0) throw new ParseError(s);
    HDate date = new HDate.fromString(s.substring(0, tIdx));

    // find timezone offset and name indices.
    // this, in turn, allows us to find the time
    if (s.endsWith("Z"))
      return new HDateTime(
          date,
          new HTime.fromString(s.substring(tIdx + 1, s.length - 1)),
          HTimeZone.UTC);
    else if (s.endsWith("Z UTC"))
      return new HDateTime(
          date,
          new HTime.fromString(
              s.substring(tIdx + 1, s.length - "Z UTC".length)),
          HTimeZone.UTC);

    // have tz offset and name
    int spIdx = s.indexOf(' ');
    if (spIdx < 0) throw new ParseError("Expected time zone name: $s");
    // work backward from space to find beginning of offset
    int offsetIdx = spIdx - 1;
    while (true) {
      if (offsetIdx <= tIdx)
        throw new ParseError("Expected Z or -/+ for timezone offset: $s");
      int c = s.codeUnitAt(offsetIdx);
      if (c == MINUS_CHAR || c == PLUS_CHAR || c == Z_CHAR) break;
      --offsetIdx;
    }
    String offsetStr = s.substring(offsetIdx, spIdx);
    int offset = "Z" == offsetStr ? 0 : _parseOffset(offsetStr);
    HTimeZone tz = new HTimeZone(s.substring(spIdx + 1), offset);

    return new HDateTime(
        date, new HTime.fromString(s.substring(tIdx + 1, offsetIdx)), tz);
  }

  factory HDateTime.now([HTimeZone tz]) =>
      new HDateTime.fromDateTime(new DateTime.now(), tz);

  factory HDateTime(HDate date, HTime time, [HTimeZone tz]) {
    if (date == null || time == null)
      throw new ArgumentError('data and time must not be null');
    if (tz == null) tz = HTimeZone.DEFAULT;
    return new HDateTime._(date, time, tz);
  }

  HDateTime._(this.date, this.time, this.tz);

  /// Date component of the timestamp
  final HDate date;

  /// Time component of the timestamp
  final HTime time;

  /// Offset in seconds from UTC including DST offset
  int get tzOffset => tz.offset;

  /// Timezone as Olson database city name
  final HTimeZone tz;

  /// Get this date time as Java milliseconds since epoch
  int millis() {
    if (_millis == null) {
      DateTime utcTime = new DateTime.utc(date.year, date.month, date.day,
          time.hour, time.min, time.sec, time.ms);
      _millis = utcTime
          .subtract(new Duration(seconds: tz.offset))
          .millisecondsSinceEpoch;
    }
    return _millis;
  }

  int _millis;

  /// Hash is based on date, time, tzOffset, and (NOT) tz
  @override
  int get hashCode {
    if (_hashCode == null) {
      _hashCode = hash3(date.hashCode, time.hashCode, tz.hashCode);
    }
    return _hashCode;
  }

  int _hashCode;

  @override
  bool operator ==(Object that) {
    return that is HDateTime &&
        date == that.date &&
        time == that.time &&
        tz == that.tz &&
        tz.name == that.tz.name;
  }

  /// Comparison based on millis.
  @override
  int compareTo(Object that) {
    int thisMillis = this.millis();
    int thatMillis = (that as HDateTime).millis();
    if (thisMillis < thatMillis)
      return -1;
    else if (thisMillis > thatMillis) return 1;
    return 0;
  }

  /// Encode as "t:YYYY-MM-DD'T'hh:mm:ss.FFFz zzzz"
  @override
  String toJson() {
    StringBuffer s = new StringBuffer();
    s.write("t:");
    _encode(s);
    return s.toString();
  }

  /// Encode as "YYYY-MM-DD'T'hh:mm:ss.FFFz zzzz"
  @override
  String toZinc() {
    StringBuffer s = new StringBuffer();
    _encode(s);
    return s.toString();
  }

  static int _parseOffset(String s) {
    if (s.length != "-HH:MM".length)
      throw new ParseError("Invalid tz offset: $s");
    int sign = s.startsWith("-") ? -1 : 1;
    int tzHours = int.parse(s.substring(1, 3));
    if (s.codeUnitAt(3) != COLON_CHAR)
      throw new ParseError("Invalid tz offset: $s");
    int tzMins = int.parse(s.substring(4));
    return sign * (tzHours * 3600) + (tzMins * 60);
  }

  void _encode(StringBuffer s) {
    date.encode(s);
    s.write('T');
    time.encode(s);
    if (tz.offset == 0)
      s.write('Z');
    else {
      int offset = this.tz.offset; //seconds
      if (offset < 0) {
        s.write('-');
        offset = -offset;
      } else {
        s.write('+');
      }

      int zh = offset ~/ 3600;
      int zm = (offset % 3600) ~/ 60;
      if (zh < 10) s.write('0');
      s..write(zh)..write(':');
      if (zm < 10) s.write('0');
      s.write(zm);
    }

    s..write(' ')..write(tz);
  }
}
