///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

part of haystack.hgrid;

/// HDict is an immutable map of name/HVal pairs.  Use HDictBuilder
/// to construct a HDict instance.
///
/// See [Project Haystack](http://project-haystack.org/doc/TagModel#tagKinds)
abstract class HDict extends HVal with MapMixin<String, HVal> {
  /// Singleton for empty set of tags.
  static final HDict EMPTY = new _MapImpl(
      new Map<String, HVal>()); //it was new MapImpl(new HashMap(11));

  factory HDict(Map<String, HVal> map) = _MapImpl;

  HDict._();

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /// Return if size is zero.
  @override
  bool get isEmpty => size == 0;

  /// Return number of tag name/value pairs
  int get size;

  /// Return if the given tag is present.
  bool has(String name) => get(name, checked: false) != null;

  /// Return if the given tag is not present.
  bool missing(String name) => get(name, checked: false) == null;

  @override
  void forEach(void action(String key, HVal value)) {
    for (String key in keys) {
      if (this[key] != null) {
        action(key, this[key]);
      }
    }
  }

  /// Get a tag by name.
  ///
  /// If not found and checked if false then
  /// return null, otherwise throw UnknownNameException.
  HVal get(String name, {bool checked: true});

  /// Create Map.Entry iteratator to walk each name/tag pair
  Iterator<DictEntry> iterator();

  /// Get the "id" tag as HRef.
  HRef id() => getRef("id");

  /// Get display string for this entity:
  ///   - dis tag
  ///   - id tag
  String dis() {
    HVal v;
    v = get("dis", checked: false);
    if (v is HStr) return v.val;

    v = get("id", checked: false);
    if (v != null) return (v as HRef).getDis();

    return "????";
  }

//////////////////////////////////////////////////////////////////////////
// Get Conveniences
//////////////////////////////////////////////////////////////////////////

  /// Get tag as HBool or raise UnknownNameException or ClassCastException.
  bool getBool(String name) => (get(name) as HBool).val;

  /// Get tag as HStr or raise UnknownNameException or ClassCastException.
  String getStr(String name) => (get(name) as HStr).val;

  /// Get tag as HRef or raise UnknownNameException or ClassCastException.
  HRef getRef(String name) => get(name) as HRef;

  /// Get tag as HNum or raise UnknownNameException or ClassCastException.
  int getInt(String name) => (get(name) as HNum).val.toInt();

  /// Get tag as HNum or raise UnknownNameException or ClassCastException.
  double getDouble(String name) => (get(name) as HNum).val;

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  /// String format is always "toZinc"
  @override
  String toString() => toZinc();

  int _hashCode;

  /// Hash code is based on tags
  @override
  int get hashCode {
    if (_hashCode == null) {
      List<DictEntry> items = new List<DictEntry>();

      for (Iterator<DictEntry> it = iterator(); it.moveNext();) {
        items.add(it.current);
      }

      _hashCode = hashObjects(items);
    }
    return _hashCode;
  }

  /// Equality is tags
  @override
  bool operator ==(Object other) {
    if (other is HDict) {
      if (size != other.size) return false;

      for (var it = iterator(); it.moveNext();) {
        DictEntry entry = it.current;
        if (entry.val != other.get(entry.key, checked: false)) return false;
      }

      return true;
    } else
      return false;
  }

//////////////////////////////////////////////////////////////////////////
// Encoding
//////////////////////////////////////////////////////////////////////////

  /// Return if the given string is a legal tag name.
  ///
  /// The first char must be ASCII lower case letter. Rest of
  /// chars must be ASCII letter, digit, or underbar.
  static bool isTagName(String n) {
    if (n.length == 0) return false;
    int first = n.codeUnitAt(0);
    if (first < a_CHAR || first > z_CHAR) return false;

    return n.codeUnits.every((int c) => c < 128 && _tagChars[c]);
  }

  static List<bool> _tagChars = () {
    List<bool> result = new List<bool>.filled(128, false);
    for (int i = a_CHAR; i <= z_CHAR; ++i) result[i] = true;
    for (int i = A_CHAR; i <= Z_CHAR; ++i) result[i] = true;
    for (int i = ZERO_CHAR; i <= NINE_CHAR; ++i) result[i] = true;
    result[UNDERSCORE_CHAR] = true;
    return result;
  }();

  /// Encode value to json format
  @override
  String toJson() {
    throw new UnsupportedError("HDict.toJson");
  }

  /// Encode value to zinc format
  @override
  String toZinc() {
    StringBuffer sb = new StringBuffer();
    sb..write('{')..write(toZincMeta(this))..write('}');
    return sb.toString();
  }

  static String toZincMeta(HDict dict) {
    StringBuffer sb = new StringBuffer();
    bool first = true;

    for (Iterator<DictEntry> it = dict.iterator(); it.moveNext();) {
      DictEntry e = it.current;
      if (first)
        first = false;
      else
        sb.write(' ');
      sb.write(e.key);
      if (e.val != HMarker.VAL) sb..write(':')..write(e.val.toZinc());
    }

    return sb.toString();
  }

  @override
  operator []=(String key, HVal value) {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }

  @override
  void clear() {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }

  @override
  HVal remove(Object key) {
    throw new UnsupportedError("Cannot modify unmodifiable map");
  }
}

//////////////////////////////////////////////////////////////////////////
// MapImpl
//////////////////////////////////////////////////////////////////////////

class _MapImpl extends HDict {
  factory _MapImpl(Map<String, HVal> map) {
    List<DictEntry> list = new List<DictEntry>();
    map.forEach((String key, HVal val) {
      list.add(new DictEntry(key, val));
    });
    return new _MapImpl._(map, list);
  }

  _MapImpl._(this._map, this._list) : super._();

  @override
  int get size => _map.length;

  @override
  HVal operator [](Object key) => get(key as String, checked: false);

  @override
  Iterable<String> get keys => _map.keys;

  @override
  HVal get(String name, {bool checked: true}) {
    HVal val = _map[name];
    if (val != null) return val;
    if (!checked) return null;
    throw new UnknownNameError(name);
  }

  @override
  Iterator<DictEntry> iterator() => _list.iterator;

  final Map<String, HVal> _map;

  final List<DictEntry> _list;
}

//////////////////////////////////////////////////////////////////////////
// MapEntry
//////////////////////////////////////////////////////////////////////////

/// Create Map.Entry for given name/value tag pair
class DictEntry {
  final String key;
  final HVal val;

  int _hashCode;
  @override
  int get hashCode =>
      (_hashCode != null) ? _hashCode : _hashCode = hash2(key, val);

  @override
  bool operator ==(Object that) =>
      that is DictEntry && that.key == key && that.val == val;

  DictEntry(this.key, this.val);
}
