///
/// Copyright (c) 2018, BAS Services & Graphics, LLC.
/// Licensed under the Academic Free License version 3.0
///

library haystack.hfilter;

import 'HBool.dart';
import 'HGrid.dart';
import 'HRef.dart';
import 'HVal.dart';
import 'ParseError.dart';
import 'io/HaystackToken.dart';
import 'io/HaystackTokenizer.dart';

/// HFilter models a parsed tag query string.
///
/// @see <a href='http://project-haystack.org/doc/Filters'>Project Haystack</a>
abstract class HFilter {
//////////////////////////////////////////////////////////////////////////
// Encoding
//////////////////////////////////////////////////////////////////////////

  /// Decode a string into a HFilter; return null or throw
  /// ParseError if not formatted correctly.
  factory HFilter.parse(String s, [bool checked = true]) {
    try {
      return new _FilterParser(s).parse();
    } catch (e) {
      if (!checked) return null;
      if (e is ParseError) throw e;
      throw new ParseError(s);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Factories
//////////////////////////////////////////////////////////////////////////

  /// Match records which have the specified tag path defined.
  static HFilter has(String path) => new _Has(new Path.parse(path));

  /// Match records which do not define the specified tag path.
  static HFilter missing(String path) => new _Missing(new Path.parse(path));

  /// Match records which have a tag are equal to the specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter eq(String path, HVal val) =>
      new _Eq(new Path.parse(path), val);

  /// Match records which have a tag not equal to the specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter ne(String path, HVal val) =>
      new _Ne(new Path.parse(path), val);

  /// Match records which have tags less than the specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter lt(String path, HVal val) =>
      new _Lt(new Path.parse(path), val);

  /// Match records which have tags less than or equals to specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter le(String path, HVal val) =>
      new _Le(new Path.parse(path), val);

  /// Match records which have tags greater than specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter gt(String path, HVal val) =>
      new _Gt(new Path.parse(path), val);

  /// Match records which have tags greater than or equal to specified value.
  ///
  /// If the path is not defined then it is unmatched.
  static HFilter ge(String path, HVal val) =>
      new _Ge(new Path.parse(path), val);

  /// Return a query which is the logical-and of this and that query.
  HFilter and(HFilter that) => new _And(this, that);

  /// Return a query which is the logical-or of this and that query.
  HFilter or(HFilter that) => new _Or(this, that);

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  HFilter._();

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /// Return if given tags entity matches this query.
  bool include(HDict dict, Pather pather);

  /// String encoding.
  @override
  String toString() {
    if (_string == null) _string = toStr();
    return _string;
  }

  String _string;

  /// Used to lazily build toString.
  String toStr();

  /// Hash code is based on string encoding.
  @override
  int get hashCode => toString().hashCode;

  /// Equality is based on string encoding.
  @override
  bool operator ==(Object that) =>
      (that is HFilter) && toString() == that.toString();
}

//////////////////////////////////////////////////////////////////////////
// HFilter.Path
//////////////////////////////////////////////////////////////////////////

/// Pather is a callback interface used to resolve query paths.
abstract class Pather {
  /// Given a HRef string identifier, resolve to an entity's
  /// HDict representation or ref is not found return null.
  HDict find(String ref);
}

//////////////////////////////////////////////////////////////////////////
// HFilter.Path
//////////////////////////////////////////////////////////////////////////

/// Path is a simple name or a complex path using the "->" separator.
abstract class Path {
  /// Construct a new Path from string or throw ParseError.
  factory Path.parse(String path) {
    try {
      // optimize for common single name case
      int dash = path.indexOf('-');
      if (dash < 0) return new Path1(path);

      // parse
      int s = 0;
      var acc = <String>[];

      while (true) {
        String n = path.substring(s, dash);
        if (n.isEmpty) throw new Exception();
        acc.add(n);
        if (path[dash + 1] != '>') throw new Exception();
        s = dash + 2;
        dash = path.indexOf('-', s);

        if (dash < 0) {
          n = path.substring(s, path.length);
          if (n.isEmpty) throw new Exception();
          acc.add(n);
          break;
        }
      }
      return new PathN(path, acc);
    } catch (e) {}

    throw new ParseError("Path: $path");
  }

  Path(); // stub for extends

  /// Number of names in the path.
  int get size;

  /// Get name at given index.
  String get(int i);

  /// Hashcode is based on string.
  @override
  int get hashCode => toString().hashCode;

  /// Equality is based on string.
  @override
  bool operator ==(Object that) => toString() == that.toString();

  /// Get string encoding.
  @override
  String toString();
}

class Path1 extends Path {
  Path1(String this.name);

  @override
  int get size => 1;

  @override
  String get(int i) {
    if (i == 0) return name;
    throw new IndexError(i, this.name);
  }

  @override
  String toString() => name;

  final String name;
}

class PathN extends Path {
  PathN(String this._string, List<String> this._names);

  @override
  int get size => _names.length;

  @override
  String get(int i) => _names[i];

  @override
  String toString() => _string;

  final String _string;
  final List<String> _names;
}

//////////////////////////////////////////////////////////////////////////
// PathFilter
//////////////////////////////////////////////////////////////////////////

abstract class PathFilter extends HFilter {
  PathFilter(Path this.path) : super._();

  @override
  bool include(HDict dict, Pather pather) {
    HVal val = dict.get(path.get(0), checked: false);

    if (path.size != 1) {
      HDict nt = dict;

      for (int i = 1; i < path.size; ++i) {
        if (val is! HRef) {
          val = null;
          break;
        }
        nt = pather.find((val as HRef).val);
        if (nt == null) {
          val = null;
          break;
        }
        val = nt.get(path.get(i), checked: false);
      }
    }

    return doInclude(val);
  }

  bool doInclude(HVal val);
  final Path path;
}

//////////////////////////////////////////////////////////////////////////
// Has
//////////////////////////////////////////////////////////////////////////

class _Has extends PathFilter {
  _Has(Path p) : super(p);

  @override
  bool doInclude(HVal v) => v != null;

  @override
  String toStr() => path.toString();
}

//////////////////////////////////////////////////////////////////////////
// Missing
//////////////////////////////////////////////////////////////////////////

class _Missing extends PathFilter {
  _Missing(Path p) : super(p);

  @override
  bool doInclude(HVal v) => v == null;

  @override
  String toStr() => "not $path";
}

//////////////////////////////////////////////////////////////////////////
// CmpFilter
//////////////////////////////////////////////////////////////////////////

abstract class _CmpFilter extends PathFilter {
  _CmpFilter(Path p, HVal this.val) : super(p);

  @override
  String toStr() {
    StringBuffer s = new StringBuffer();
    s..write(path)..write(cmpStr())..write(val.toZinc());
    return s.toString();
  }

  bool sameType(HVal v) => v != null && v.runtimeType == val.runtimeType;
  String cmpStr();
  final HVal val;
}

//////////////////////////////////////////////////////////////////////////
// Eq
//////////////////////////////////////////////////////////////////////////

class _Eq extends _CmpFilter {
  _Eq(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => "==";

  @override
  bool doInclude(HVal v) => v != null && v == val;
}

//////////////////////////////////////////////////////////////////////////
// Ne
//////////////////////////////////////////////////////////////////////////

class _Ne extends _CmpFilter {
  _Ne(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => "!=";

  @override
  bool doInclude(HVal v) => v != null && v != val;
}

//////////////////////////////////////////////////////////////////////////
// Lt
//////////////////////////////////////////////////////////////////////////

class _Lt extends _CmpFilter {
  _Lt(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => "<";

  @override
  bool doInclude(HVal v) => sameType(v) && v.compareTo(val) < 0;
}

//////////////////////////////////////////////////////////////////////////
// Le
//////////////////////////////////////////////////////////////////////////

class _Le extends _CmpFilter {
  _Le(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => "<=";

  @override
  bool doInclude(HVal v) => sameType(v) && v.compareTo(val) <= 0;
}

//////////////////////////////////////////////////////////////////////////
// Gt
//////////////////////////////////////////////////////////////////////////

class _Gt extends _CmpFilter {
  _Gt(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => ">";

  @override
  bool doInclude(HVal v) => sameType(v) && v.compareTo(val) > 0;
}

//////////////////////////////////////////////////////////////////////////
// Ge
//////////////////////////////////////////////////////////////////////////

class _Ge extends _CmpFilter {
  _Ge(Path p, HVal v) : super(p, v);

  @override
  String cmpStr() => ">=";

  @override
  bool doInclude(HVal v) => sameType(v) && v.compareTo(val) >= 0;
}

//////////////////////////////////////////////////////////////////////////
// Compound
//////////////////////////////////////////////////////////////////////////

abstract class _CompoundFilter extends HFilter {
  _CompoundFilter(HFilter this.a, HFilter this.b) : super._();

  String get keyword;

  @override
  String toStr() {
    StringBuffer s = new StringBuffer();
    if (a is _CompoundFilter)
      s..write('(')..write(a)..write(')');
    else
      s.write(a);
    s..write(' ')..write(keyword)..write(' ');
    if (b is _CompoundFilter)
      s..write('(')..write(b)..write(')');
    else
      s.write(b);
    return s.toString();
  }

  final HFilter a;
  final HFilter b;
}

//////////////////////////////////////////////////////////////////////////
// And
//////////////////////////////////////////////////////////////////////////

class _And extends _CompoundFilter {
  _And(HFilter a, HFilter b) : super(a, b);

  @override
  String get keyword => "and";

  @override
  bool include(HDict dict, Pather pather) {
    return a.include(dict, pather) && b.include(dict, pather);
  }
}

//////////////////////////////////////////////////////////////////////////
// Or
//////////////////////////////////////////////////////////////////////////

class _Or extends _CompoundFilter {
  _Or(HFilter a, HFilter b) : super(a, b);

  @override
  String get keyword => "or";

  @override
  bool include(HDict dict, Pather pather) =>
      a.include(dict, pather) || b.include(dict, pather);
}

//////////////////////////////////////////////////////////////////////////
// FilterParser
//////////////////////////////////////////////////////////////////////////

class _FilterParser {
  _FilterParser(String str) {
    this._tokenizer = new HaystackTokenizer(str.runes);
    _consume();
    _consume();
  }

  HFilter parse() {
    HFilter f = _condOr();
    _verify(HaystackToken.eof);
    return f;
  }

  HFilter _condOr() {
    HFilter lhs = _condAnd();
    if (!_isKeyword("or")) return lhs;
    _consume();
    return lhs.or(_condOr());
  }

  HFilter _condAnd() {
    HFilter lhs = _term();
    if (!_isKeyword("and")) return lhs;
    _consume();
    return lhs.and(_condAnd());
  }

  HFilter _term() {
    if (_cur == HaystackToken.lparen) {
      _consume();
      HFilter f = _condOr();
      _consume(HaystackToken.rparen);
      return f;
    }

    if (_isKeyword("not") && _peek == HaystackToken.id) {
      _consume();
      return new _Missing(_path());
    }

    Path p = _path();
    if (_cur == HaystackToken.eq) {
      _consume();
      return new _Eq(p, _val());
    }
    if (_cur == HaystackToken.notEq) {
      _consume();
      return new _Ne(p, _val());
    }
    if (_cur == HaystackToken.lt) {
      _consume();
      return new _Lt(p, _val());
    }
    if (_cur == HaystackToken.ltEq) {
      _consume();
      return new _Le(p, _val());
    }
    if (_cur == HaystackToken.gt) {
      _consume();
      return new _Gt(p, _val());
    }
    if (_cur == HaystackToken.gtEq) {
      _consume();
      return new _Ge(p, _val());
    }

    return new _Has(p);
  }

  Path _path() {
    String id = pathName();
    if (_cur != HaystackToken.arrow) return new Path1(id);

    var segments = <String>[id];
    StringBuffer s = new StringBuffer()..write(id);

    while (_cur == HaystackToken.arrow) {
      _consume(HaystackToken.arrow);
      id = pathName();
      segments.add(id);
      s..write(HaystackToken.arrow)..write(id);
    }
    return new PathN(s.toString(), segments);
  }

  String pathName() {
    if (_cur != HaystackToken.id)
      throw _err("Expecting tag name, not " + _curToStr());
    String id = _curVal as String;
    _consume();
    return id;
  }

  HVal _val() {
    if (_cur.literal) {
      HVal val = _curVal as HVal;
      _consume();
      return val;
    }

    if (_cur == HaystackToken.id) {
      if ("true" == _curVal) {
        _consume();
        return HBool.TRUE;
      }
      if ("false" == _curVal) {
        _consume();
        return HBool.FALSE;
      }
    }

    throw _err("Expecting value literal, not " + _curToStr());
  }

  bool _isKeyword(String n) {
    return _cur == HaystackToken.id && n == _curVal;
  }

  void _verify(HaystackToken expected) {
    if (_cur != expected) throw _err("Expected $expected not " + _curToStr());
  }

  String _curToStr() => _curVal != null ? "$_cur $_curVal" : _cur.toString();

  void _consume([HaystackToken expected]) {
    if (expected != null) _verify(expected);
    _cur = _peek;
    _curVal = _peekVal;
    _peek = _tokenizer.next();
    _peekVal = _tokenizer.val;
  }

  ParseError _err(String msg) => new ParseError(msg);

  HaystackTokenizer _tokenizer;
  HaystackToken _cur;
  Object _curVal;
  HaystackToken _peek;
  Object _peekVal;
}
